<?php

/**
 * This file add most wordpress hooks to catch
 * events that matter to us.
 *
 * @link       https://crowdaa.com
 * @since      1.0.0
 *
 * @package    Crowdaa-Sync
 * @subpackage Crowdaa-Sync/admin
 * @author     Crowdaa <contact@crowdaa.com>
 */
class Crowdaa_Sync_WP_Hooks {
  /**
   * Initialize the class and set its properties.
   *
   * @since    1.0.0
   */
  public function __construct() {
    add_action('save_post', [$this, 'save_post_without_cat_notice']);
    add_action('save_post', [$this, 'set_need_to_sync_for_post']);
    add_action('admin_notices', [$this, 'sync_admin_notice']);
    add_action('before_delete_post', [$this, 'sync_delete_post']);
    add_shortcode('display_gallery', [$this, 'display_gallery']);
  }

  /**
   * Show admin notice
   *
   * @since    1.0.0
   * @param $post_id
   * @return void
   */
  public function save_post_without_cat_notice($post_id) {
    // if autosave or revision don't continue
    if (wp_is_post_autosave($post_id) || wp_is_post_revision($post_id)) {
      return;
    }

    if (isset($_POST['post_type']) && 'post' != $_POST['post_type']) {
      return;
    }

    $post_type_taxonomy = 'category';
    $terms = get_the_terms($post_id, $post_type_taxonomy);
    $gallery = get_post_meta($post_id, 'second_featured_img', true);

    if (empty($terms)) {
      $this->sync_admin_notice();
    }

    if (empty($gallery)) {
      $this->sync_admin_notice();
    }
  }

  /**
   * Set need sync option for synchronization
   *
   * @since    1.0.0
   * @return void
   */
  public function sync_admin_notice() {
    global $pagenow;
    global $post;
    if (($pagenow == 'post.php') && ($post->post_type == 'post')) {
      $post_type_taxonomy = 'category';
      $terms = get_the_terms(get_the_ID(), $post_type_taxonomy);
      if (empty($terms)) {
        Crowdaa_Sync_Admin_Display::admin_notice('error', __('Crowdaa-Sync: You will not be able to synchronize this post until you add a Sync category', CROWDAA_SYNC_PLUGIN_NAME));
      }
    }
  }

  /**
   * Set need sync option for post synchronization
   *
   * @since    1.0.0
   * @param $post_id
   * @return void
   */
  public function set_need_to_sync_for_post($post_id) {
    update_post_meta($post_id, 'crowdaa_need_sync', 'yes');
  }

  /**
   * Hide deleted post from WP to API
   *
   * @since    1.0.0
   * @param $post_id
   * @return void
   */
  public function sync_delete_post($post_id) {
    $post_api_id = get_post_meta($post_id, 'api_post_id', true);
    $sync_api    = new Crowdaa_Sync_API();
    if ($post_api_id) {
      $sync_api->delete_post_api($post_api_id);
    }
  }

  /**
   * Shortcode to display gallery in an article
   *
   * @since    1.0.0
   * @return string
   */
  function display_gallery() {
    $html          = '<div class="crowdaa_gallery">';
    $pictures_list = get_post_meta(get_the_ID(), 'second_featured_img', true);
    if ($pictures_list) {
      $ids = explode(',', $pictures_list);
      foreach ($ids as $attachment_id) {
        $img = wp_get_attachment_image_src($attachment_id, 'thumbnail');
        if ($img) {
          $html .= '<div class="screen-thumb"><img src="'.esc_url($img[0]).'" /></div>';
        } else {
          $video_url = wp_get_attachment_url($attachment_id);
          $html     .= '<div class="screen-thumb"><video width="200" height="150" controls="controls">
                          <source src="'.$video_url.'" type="video/mp4">
                        </video></div>';
        }
      }
      return $html.'</div>';
    }
  }
}
